/*
 * gb2u.c - routines for converting gb code to unicode
 *
 * Cheng Lei  2000.6.6
 */ 

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "gb2u.h"

#define MAX_TABLE_LEN 		65536
#define MAX_FILE_LEN            4096*1024
#define MAX_LINE_LEN		1024		

/*
 *  gbTable [gb code] = unicode
 */
int gb2u_table[MAX_TABLE_LEN] = {0};


/**** blow is inner function declare ***/
/*
   read the whole file into mem. if file length > MAX_FILE_LEN,function return 
   -2.if file not exist ,return -1.other error return -3.
   function return 0 if success.
   function will malloc mem for buf,after using buf,don't forget free it.
*/
int g2u_get_file(const char* filename,char** buf,int* buf_len)
{
	FILE *f;
	int  file_len = 0,file_head_len = 0,file_end_len = 0;
	
	*buf_len = 0;
	*buf = NULL;
   	f = fopen(filename,"r");
   	if (f == NULL)
   	{
		printf("fopen failed can't open file:%s.",filename);
        	return -1;
   	}
   
   	fseek(f,0,SEEK_END);
   	file_end_len = ftell(f);
   	fseek(f,0,SEEK_SET);
   	file_head_len = ftell(f);
 	file_len = file_end_len - file_head_len;

	if (file_len > MAX_FILE_LEN)
   	{
		fclose(f);
		printf("file gbTounicode exceeded max length.");
		return -2;
   	}
   
   	*buf = malloc(file_len + 1);
   	if (*buf == NULL)
   	{
		fclose(f);
       		printf("malloc error.");
       		return -3;
   	}

   	fread(*buf,sizeof(char),file_len,f);
   	fclose(f);
	*buf_len = file_len;
	(*buf)[file_len] = 0;
	return 0;
}


int g2u_init(const char* filename)
{
   	char *buf,*cur_pos,*nxt_pos = NULL;
	char is_continue = 1;
   	int  buf_len = 0;
   	int  gb,uni,ret;
        
   	if ((ret = g2u_get_file(filename,&buf,&buf_len)) < 0)
	{
		return ret;
	} 
	
	cur_pos = buf;
   	while (is_continue)
	{
		nxt_pos = strchr(cur_pos,'\n');
		if (nxt_pos == NULL)
		{
			is_continue = 0;
		}
		
		if (sscanf(cur_pos + 2," %X %X",&gb,&uni) == 2);
		{
			// A0A0ʼFEFE,ÿһԪطһӦunicodeֵ
			// ʵunicode.orgϵĶԻ,F7FE
			if (gb < MAX_TABLE_LEN) 
				gb2u_table[gb] = uni;
		}
		cur_pos = nxt_pos + 1;		
   	}
	
	free(buf);
   	return 0;
}


int g2u_convert_unicode(char* gbchar,int gb_len,char* unichar,int uni_len)
{
	int num = 0;
	int gb_pos = 0,uni_pos = 0;
	int ret = 0;

	if ((gbchar == NULL) || (unichar == NULL) || (gb_len <= 0) || (uni_len <= 0))
		return 0;
	
	while ((gbchar[gb_pos] != 0) && (gb_pos < gb_len) && (uni_pos < uni_len))
	{		
		
		// һַ>=A0,һַҲ>=A0,Ϊһ(Gb2unicodeԻ
		// ) һֵķΧΪ(c1>0xa0 && c1<0xff) and (c2>0xa0 && c2<0xff) 
		if (((unsigned char)gbchar[gb_pos] >= 0xA0) && ((unsigned char)gbchar[gb_pos + 1] >= 0xA0)
				&& (gb_pos + 1 < gb_len) && (uni_pos + 1 < uni_len))
		{
			// һַ*256(8bit)ϵڶַõһֵhexֵ,
			// ȥõunicodeֵ
			num = ((unsigned char)gbchar[gb_pos])*256 + ((unsigned char)gbchar[gb_pos + 1]);		
			if (num == 0) // ?
			{
			 	unichar[uni_pos++] = gbchar[gb_pos++];
			 	unichar[uni_pos++] = gbchar[gb_pos++];
			}
			else // õunicodeֵ֮,ٷֱȡ8bit͵8bitchar
			{
				unichar[uni_pos++] = gb2u_table[num] / 256;
				unichar[uni_pos++] = gb2u_table[num] % 256;
				gb_pos += 2;					
				ret ++;
			}
		}
		else // gb2312ַ--ascii
		{
			unichar[uni_pos++] = gbchar[gb_pos++];
		}
			 	
	}
	return ret;
}

int g2u_convert_utf8(char* gbchar,int gb_len,char* utfchar,int utf_len,int* new_len)
{
	int num = 0;
	int gb_pos = 0,utf_pos = 0;
	unsigned char uni_code[2];
	int ret = 0;

	if ((gbchar == NULL) || (utfchar == NULL) || (gb_len <= 0) || (utf_len <= 0))
	{
		*new_len = gb_len;
		return 0;
	}
	
	while ((gbchar[gb_pos] != 0) && (gb_pos < gb_len) && (utf_pos < utf_len))
	{		
		
		if (((unsigned char)gbchar[gb_pos] >= 0xA0) && ((unsigned char)gbchar[gb_pos + 1] >= 0xA0)
				&& (gb_pos + 1 < gb_len) && (utf_pos + 1 < utf_len))
		{
			num = ((unsigned char)gbchar[gb_pos])*256 + ((unsigned char)gbchar[gb_pos + 1]);		
			if (num == 0)
			{
			 	utfchar[utf_pos++] = gbchar[gb_pos++];
			 	utfchar[utf_pos++] = gbchar[gb_pos++];
			}
			else
			{
				/* unicode -> utf8 ת
				U-00000000 - U-0000007F:  0xxxxxxx  
				U-00000080 - U-000007FF:  110xxxxx 10xxxxxx  
				U-00000800 - U-0000FFFF:  1110xxxx 10xxxxxx 10xxxxxx  
				U-00010000 - U-001FFFFF:  11110xxx 10xxxxxx 10xxxxxx 10xxxxxx  
				U-00200000 - U-03FFFFFF:  111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx  
				U-04000000 - U-7FFFFFFF:  1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx  
				*/
				/*
				// ׼㷨
				if (gb2u_table[num] < 0x80) { 
					utfchar[utf_pos++]=gb2u_table[num]; 
				}
				else if (gb2u_table[num] < 0x800) { 
					// unicode >= 0x80 and < 0x7ff,unicode6bitõߵ5bit,
					// ֮0xc0,Ӷõutf8ĸ8bit
					// Ȼunicode6bit(Ǹղߵ)0x80
					// õutf8ĵ8bit
					utfchar[utf_pos++]=(0xC0 | gb2u_table[num]>>6); 
					utfchar[utf_pos++]=(0x80 | gb2u_table[num] & 0x3F); 
				} 
				else if (gb2u_table[num] < 0x10000) { 
					// ReComments!: &ȼ| ,˳Ǵҵ
					// unicode >= 0x800 and < 0xffff,unicode12bitõߵ4bit,
					// ֮0xe0,Ӷõutf8utf8bit
					// Ȼ6bit0x3f,ٺ0x80,õutfм8bit,
					// unicode0x3f֮0x80,õutf͵8bit
					utfchar[utf_pos++]=(0xE0 | gb2u_table[num]>>12); 
					utfchar[utf_pos++]=(0x80 | gb2u_table[num]>>6 & 0x3F); 
					utfchar[utf_pos++]=(0x80 | gb2u_table[num] & 0x3F); 
				} 
				else if (gb2u_table[num] < 0x200000) { 
					utfchar[utf_pos++]=(0xF0 | gb2u_table[num]>>18); 
					utfchar[utf_pos++]=(0x80 | gb2u_table[num]>>12 & 0x3F); 
					utfchar[utf_pos++]=(0x80 | gb2u_table[num]>>6 & 0x3F); 
					utfchar[utf_pos++]=(0x80 | gb2u_table[num] & 0x3F); 
				}
				*/
				// 㷨,Ϊgbֵ 0x800 < gb < 0xffff֮
				/*
				uni_code[0] = (gb2u_table[num] / 256); // unicode 8bit
				uni_code[1] = (gb2u_table[num] % 256); // unicode 8bit
				// ȡλǰ4bit0xe0,õutfߵ8bit
				utfchar[utf_pos++] = ( uni_code[0] >> 4) | 0xE0;
				// λеĵ4bit2bitλunicodeĸ2bit,֮0x80
				// 0xbfһ,Ա֤7bit0
				utfchar[utf_pos++] = ((( uni_code[0] & 0x0f) * 4) + (uni_code[1] >> 6) + 0x80) & 0xBF;
				utfchar[utf_pos++] = (uni_code[1] | 0x80) & 0xBF;
				*/
				// 㷨,ISO/IEC JTC1/SC2/WG2 N 1335
				utfchar[utf_pos++] = 0xE0 + gb2u_table[num]/(int)pow(2,12);
				utfchar[utf_pos++] = 0x80 + gb2u_table[num]/(int)pow(2,6)%(int)pow(2,6); // x%40 == x & 3f
				utfchar[utf_pos++] = 0x80 + gb2u_table[num]%(int)pow(2,6);

				gb_pos += 2;					
				ret ++;
			}
		}
		else
		{
			utfchar[utf_pos++] = gbchar[gb_pos++];
		}
			 	
	}
	*new_len = utf_pos;
	return ret;
}


int g2u_convert_utf16(char* gbchar,int gb_len,char* utfchar,int utf_len,int* new_len)
{
	int num = 0;
	int gb_pos = 0,utf_pos = 0;
	unsigned char uni_code[2];
	int ret = 0;

	if ((gbchar == NULL) || (utfchar == NULL) || (gb_len <= 0) || (utf_len <= 0))
	{
		*new_len = gb_len;
		return 0;
	}
	
	while ((gbchar[gb_pos] != 0) && (gb_pos < gb_len) && (utf_pos < utf_len))
	{		
		
		if (((unsigned char)gbchar[gb_pos] >= 0xA0) && ((unsigned char)gbchar[gb_pos + 1] >= 0xA0)
				&& (gb_pos + 1 < gb_len) && (utf_pos + 1 < utf_len))
		{
			num = ((unsigned char)gbchar[gb_pos])*256 + ((unsigned char)gbchar[gb_pos + 1]);		
			if (num == 0)
			{
			 	utfchar[utf_pos++] = gbchar[gb_pos++];
			 	utfchar[utf_pos++] = gbchar[gb_pos++];
			}
			else
			{
				/* 
				   Encoding UTF-16 --- RFC 2781  
				   Encoding of a single character from an ISO 10646 character value to
				   UTF-16 proceeds as follows. Let U be the character number, no greater
				   than 0x10FFFF.
				   1) If U < 0x10000, encode U as a 16-bit unsigned integer and
					  terminate.
				   2) Let U' = U - 0x10000. Because U is less than or equal to 0x10FFFF,
					  U' must be less than or equal to 0xFFFFF. That is, U' can be
					  represented in 20 bits.
				   3) Initialize two 16-bit unsigned integers, W1 and W2, to 0xD800 and
					  0xDC00, respectively. These integers each have 10 bits free to
					  encode the character value, for a total of 20 bits.
				   4) Assign the 10 high-order bits of the 20-bit U' to the 10 low-order
					  bits of W1 and the 10 low-order bits of U' to the 10 low-order
					  bits of W2. Terminate.
				*/
				if (gb2u_table[num] < 0x10000) 
				{ 
					utfchar[utf_pos++] = (gb2u_table[num]>>8) & 0xff; 
					utfchar[utf_pos++] = gb2u_table[num] & 0xff; 
				}else
				{
					// Unicode Transformation Formats: UTF-8 & Co.
					// unicode 16bit,d7c0unicode6bit
					utfchar[utf_pos++] = (0xD7C0 + (gb2u_table[num] >> 10)); 
					utfchar[utf_pos++] = (0xDC00 | gb2u_table[num] & 0x3FF);
					
					/* rfc
					unsigned int  w1, w2; 
					long int tu ;// template utf16 code
					long int bigcode;// template input chart
					tu = bigcode - 0x10000;
					w1 = 0xD800 | tu>>10 & 0x3ff; 
					w2 = 0xDC00 | tu & 0x3ff;
					*/
				}
				gb_pos += 2;					
				ret ++;
			}
		}
		else
		{
			utfchar[utf_pos++] = gbchar[gb_pos++];
		}
			 	
	}
	*new_len = utf_pos;
	return ret;
}


int main(int argc,char* argv[])
{
	FILE *f;
	char *file_content = NULL;
	char *new_content = NULL;
	int file_len,file_begin,file_end,new_len;
	
	if (argc < 4)
	{
		printf("Usage: 	%s <code table file> <source file> <destination file> \n", argv[0]);
		printf("  <code table file> : chinese and unicode code reference table.\n");
		printf("  <source file>     : file contain the chinese word you want to convert.\n");
		printf("  <destination file>: converted result file.\n");
		return -1;
	}
		
	if (g2u_init(argv[1]) != 0)
	{
		printf("Init g2u error.");
		return -2;
	}
	
	if ((f = fopen(argv[2],"r")) == NULL)
	{
        	printf("source file can't be opened.");
		return -3;	
	}
	
	fseek(f,0,SEEK_END);
	file_end = ftell(f);
	fseek(f,0,SEEK_SET);
	file_begin = ftell(f);
	file_len = file_end - file_begin;
	
	file_content = malloc(file_len + 1);
	if (file_content == NULL)
	{
		printf("fail to malloc mem.");
		return -4;	
	}
	
	fread(file_content,sizeof(char),file_len,f);
	fclose(f);
	
	file_content[file_len] = 0;
	
	new_content = malloc(file_len * 2);
	if (new_content == NULL)
		return -6;
	
	//g2u_convert_utf8(file_content,file_len,new_content,file_len*2,&new_len);		
	g2u_convert_utf16(file_content,file_len,new_content,file_len*2,&new_len);		
	
	if ((f = fopen(argv[3],"w")) == NULL)
	{
		free(file_content);
		printf("result file can't be opened.");
		return -5;
	}
	
	fwrite(new_content,sizeof(char),new_len,f);
	fclose(f);
	free(file_content);
	
	
	return 0;
	
}




