/*
*  gb2u.h - declarations for functions that parse a gb2312 code to unicode
*
*  chenglei 
*  2000.6.6
*/

#ifndef GB2U_H
#define GB2U_H

/*  
*   PURPOSE:
*    	this function setup the code table,use this table gb2unicode 
*       can parse a gb chinese character to a unicode. Before other 
*	module call the gb2unicode function , this function must be 
*	called.
*   PARAMETER:
*       filename - is the code table file;
*   RETURN:
*       0 - success,1 - file not found, 2 - other error
*/
int g2u_init(const char* filename);

/*
*   PURPOSE:
*       function convert gb code to unicode. you give the 
*       gbchar string and string length,all the gb character
*	in the string will be changed to unicode.
*	convertion result put into unichar ,if result string length > uni_len
*	procedure break and return.
*   PARAMETER:
*       gbchar - is gb encoding string.
*       gb_len  - length of gbchar to be converted,counted as bytes.
* 	unichar - put result of convertion into unichar, malloc by caller ,its length must not < the gbchar strign length.
*	uni_len - length of unichar ,counted as bytes.
*   RETURN:
*	return the number of converted chinese word.
*/

int g2u_convert_unicode(char* gbchar,int gb_len,char* unichar,int uni_len);


/*
*   PURPOSE:
*       function convert gb code to utf8. you give the
*       gbchar string and string length,all the gb character
*	in the string will be changed to utf8.
*	convertion result put into utfchar ,if result string length > utf_len
*	procedure break and return.
*   PARAMETER:
*       gbchar - is gb encoding string.
*       gb_len  - length of gbchar to be converted,counted as bytes.
* 	utfchar - put result of convertion into utfchar, malloc by caller ,
	          utfchar size should be 1 and 1/2 of the size of gbchar at least.
*	utf_len - length of utfchar ,counted as bytes.
*	new_len - contain the value of new utf string length.
*   RETURN:
*	return the number of converted chinese word.
*/

int g2u_convert_utf8(char* gbchar,int gb_len,char* utfchar,int utf_len,int* new_len);

#endif


